/*! ********************************************
 * GrayBoxSound.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox 
{
import flash.events.SampleDataEvent;
import flash.media.Sound;
import flash.media.SoundChannel;
import flash.utils.ByteArray;
public final class GrayBoxSound 
{
	private const _frameTime  :Number = 1000.0 / 44100.0;
	// object
	private var _grayblox     :GrayBox = null;
	private var _sound        :Sound = null;
	private var _soundChannel :SoundChannel = null;
	public function get soundChannel():SoundChannel { return _soundChannel; }
	// AudioBuffer
	private var _audioBufferSize:int = 4096;
	private var _audioBuffer	:Vector.<Vector.<Number>> = null;
	// status
	private var _pause :Boolean = true;
// buffer -----
	private function _clearAudioBuffer():void
	{
		var _bufferL:Vector.<Number> = _audioBuffer[0],
			_bufferR:Vector.<Number> = _audioBuffer[1];
		for (var i:int = 0; i < _audioBufferSize; i++)
		{
			_bufferL[i] = 0.0;
			_bufferR[i] = 0.0;
		}
	}
	public function getAudioBufferSize():int{ return _audioBufferSize; }
	public function setAudioBufferSize(size:int):int
	{
		if (size == _audioBufferSize) return _audioBufferSize;
		if (size != 2048 && 
			size != 4096 && 
			size != 8192)size = 4096;
		_audioBufferSize = size;
		_audioBuffer[0].fixed = false;
		_audioBuffer[1].fixed = false;
		_audioBuffer[0].length = size;
		_audioBuffer[1].length = size;
		_audioBuffer[0].fixed = true;
		_audioBuffer[1].fixed = true;
		_clearAudioBuffer();
		_grayblox.setProcessBlockSize(_audioBufferSize);
		return _audioBufferSize;
	}
// status ---
	public function isPlay():Boolean { return _soundChannel != null; }
	public function isPause():Boolean { return _pause; }
	//public function getSampleRate():Number{ return _grayblox.getSampleRate(); }
	//public function setSampleRate(sampleRate:Number):Number
	//{
		//var _sampleRate:Number = _grayblox.setSampleRate(sampleRate);
		//_frameTime = 1.0 / _sampleRate;
		//return _sampleRate;
	//}
// Main System ---
	public function initialize():void
	{
		stop();
		_clearAudioBuffer();
		_grayblox.initialize();
	}
// Module ---
	public function getModuleSlots():int { return _grayblox.getModuleSlots(); }
	public function setModuleSlots(num:int):int { return _grayblox.setModuleSlots(num); }
	public function getModule(index:int):IGrayBoxModule { return _grayblox.getModule(index); }
	public function attachModule(index:int, module:IGrayBoxModule):IGrayBoxModule { return _grayblox.attachModule(index, module); }
	public function detachModule(index:int):IGrayBoxModule { return _grayblox.detachModule(index); }
	public function resetModules():void { _grayblox.resetModules(); }
	public function initiModules():void { _grayblox.initiModules(); }
	public function getBypassStatus(index:int):Boolean { return _grayblox.getBypassStatus(index); }
	public function bypassProcess(index:int, bypass:Boolean):void { _grayblox.bypassProcess(index, bypass); }
// control ---
	public function play():SoundChannel
	{
		if (_soundChannel != null) return _soundChannel;
		_pause = false;
		// graybox
		_grayblox.startProcess();
		// sound
		if (_sound.hasEventListener(SampleDataEvent.SAMPLE_DATA))_sound.removeEventListener(SampleDataEvent.SAMPLE_DATA, _sampleDataEvent);
		_sound.addEventListener(SampleDataEvent.SAMPLE_DATA, _sampleDataEvent);
		_soundChannel = _sound.play();
		// If no SoundDriver exists
		if (_soundChannel == null) 
		{
			_sound.removeEventListener(SampleDataEvent.SAMPLE_DATA, _sampleDataEvent);
			return null;
		}
		return _soundChannel;
	}
	public function stop():void
	{
		if (_soundChannel == null) return;
		_pause = true;
		// graybox
		_grayblox.stopProcess();
		// Sound
		_soundChannel.stop();
		_sound.removeEventListener(SampleDataEvent.SAMPLE_DATA, _sampleDataEvent);
		_soundChannel = null;
	}
	public function resume():void
	{
		if (_soundChannel == null) return;
		if (_pause)_grayblox.resumeProcess();
		_pause = false;
	}
	public function pause():void
	{
		if (_soundChannel == null) return;
		if (!_pause)_grayblox.suspendProcess();
		_pause = true;
	}
// Process ---
	private function _sampleDataEvent(event:SampleDataEvent):void
	{
		var i:int, _size:int = _audioBufferSize, _stream:ByteArray = event.data;
		if (_pause) for (i = 0; i < _size; i++)_stream.writeDouble(0.0);
		else 
		{
			_clearAudioBuffer();
			if (_soundChannel != null)_grayblox.audioLatency(event.position * _frameTime - _soundChannel.position);
			_grayblox.process(_audioBuffer, 0, _size);
			var _bufferL:Vector.<Number> = _audioBuffer[0],
				_bufferR:Vector.<Number> = _audioBuffer[1];
			for (i = 0; i < _size; i++)
			{
				_stream.writeFloat(_bufferL[i]);
				_stream.writeFloat(_bufferR[i]);
			}
		}
	}
	// Constructor
	public function GrayBoxSound(audioBufferSize:int = 4096, modules:int = 4):void 
	{
		_sound = new Sound();
		if (audioBufferSize != 2048 && audioBufferSize != 2048 && 
			audioBufferSize != 4096 && 
			audioBufferSize != 8192)audioBufferSize = 4096;
		_audioBufferSize = audioBufferSize;
		_grayblox = new GrayBox(audioBufferSize, 44100.0, 2, modules);
		_audioBuffer = new Vector.<Vector.<Number>>(2, true);
		_audioBuffer[0] = new Vector.<Number>(audioBufferSize, true);
		_audioBuffer[1] = new Vector.<Number>(audioBufferSize, true);
	}
}
}