/*! ********************************************
 * SMFSequence.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.smf
{
import flash.utils.ByteArray;
import flash.utils.Endian;
public class SMFSequence
{
// Instance -----
	private var _format:int = 0;
	private var _multiPacketSysEx:Boolean = true;
	
	private var _track :SMFTrackEvent = null;  // Event List - "MTrk" Events
	public var division :int = 480;  // "MThd" divison
	// Arbitrary Loop
	private var _loopStart :SMFTrackEvent = null;
	private var _loopEnd :SMFTrackEvent = null;
// getter, setter ---
	public function setTrack(track:SMFTrackEvent):void
	{
		clear();
		_track = track;
	}
	public function getTrack():SMFTrackEvent{ return _track; }
	public function getLoopable():Boolean{ return _loopStart != null; }
	public function getLoopStart():SMFTrackEvent{ return _loopStart; }
	public function getLoopEnd():SMFTrackEvent{ return _loopEnd; }
	public function setLoopPoint(start:SMFTrackEvent = null, end:SMFTrackEvent = null):Boolean
	{
		if (_track == null) return false;
		_loopStart = null;
		_loopEnd = null;
		if (start == null || end == null || start == end) return false;
		var _list:SMFTrackEvent = start, _deltaTimes:int = 0;
		// check -->
		while (_list != null)
		{
			if (_list == end)
			{
				if (_deltaTimes == 0) return false;
				_loopStart = start;
				_loopEnd = end;
				return true;
			}
			_deltaTimes += _list.deltaTime;
			_list = _list.next;
		}
		return false;
	}
	public function clear():void
	{
		_track = null;
		division = 480;
		_loopStart = null;
		_loopEnd = null;
	}
// Constructor ---
	public function SMFSequence(bytes:ByteArray = null, multiPacketSysEx:Boolean = true):void
	{
		if (bytes != null) loadBytes(bytes, multiPacketSysEx);
	}
	public function loadBytes(bytes:ByteArray, multiPacketSysEx:Boolean = true):Boolean
	{
		var _pos:int = bytes.position,
			_endi:String = bytes.endian;
		bytes.endian = Endian.BIG_ENDIAN;
	// Start -->
		// 'MThd' ---
		if (bytes.bytesAvailable <= 16) return false;
		_multiPacketSysEx = multiPacketSysEx;
		if (bytes.readMultiByte(4, 'us-ascii') != 'MThd') return false;
		bytes.position += 4;  // headerSize
		_format	= bytes.readUnsignedShort();
		bytes.position += 2;  // trks
		var _div:int = bytes.readUnsignedShort();
		if (_format >= 2 || _div & 0x8000) return false;  // No Suport "Format2", "SMPTE"
		// 'MTrk' ---
		var _chunkType:String, _chunkDataSize:uint,
			_head:SMFTrackEvent = new SMFTrackEvent(0, 0x80, 0, 0); // dummy
		while (bytes.bytesAvailable >= 8)
		{
			_chunkType = bytes.readMultiByte(4, 'us-ascii');
			_chunkDataSize = bytes.readUnsignedInt();
			switch (_chunkType)
			{
				default     : bytes.position += _chunkDataSize; break;  // Unknown Chunk
				case 'MTrk' : 
					if (_read_MTrk(bytes, _chunkDataSize, _head)) break;
					else return false;
			}
		}
	// SetUp -->
		clear();
		_track = _head.next;
		if (_format != 0)_removeUnecEvents(_track);
		division = _div;
	// <-- Finish
		bytes.position = _pos;
		bytes.endian = _endi;
		return true;
	}
	private function _read_MTrk(bytes:ByteArray, chunkDataSize:int, trk:SMFTrackEvent):Boolean
	{
		var  _chunkEnd:int = bytes.position + chunkDataSize;
		
		var _status:int = 0, _eventType:int = 0, _channel:int = 0;
		var _val:int, _deltaTime:int = 0, _skipTime:int = 0;
		var _event:SMFTrackEvent = null;
		var _sysExDataSize:int, _sysExPacketsSize:int;
		var _metaType:int, _metaDataSize:int;
		//var _sysExTotalSize:int;
		
		// Start -->
		while (bytes.position < _chunkEnd)
		{
		// Read DeltaTime ---
			_deltaTime = _readVarLeng(bytes);
			if (_skipTime > 0){ _deltaTime += _skipTime; _skipTime = 0; }
		// Read StatusByte ---
			_val = bytes.readByte() & 0xff;
			if (_val & 0x80) // update status
			{
				_status = _val;
				if (_status < 0xF0) { _eventType = _status & 0xf0; _channel = _status & 0xf; }  // Channel Message
				else { _eventType = _status; _channel = 0; }  // reset Status, SysEx, MetaEvent
			}
			else bytes.position--;  // Runnig Status
			
			switch (_eventType)
			{
				default : bytes.position = _chunkEnd; return false; // unknown
			// MIDI Channel Message ---
			case 0x80 : case 0x90 : case 0xA0 : case 0xB0 : case 0xE0 :
				_event = new SMFTrackEvent(_deltaTime, _eventType, _channel, bytes.readByte(), bytes.readByte());
				trk = _insertTrackEvent(trk, _event);
				continue;
				
			case 0xC0 : case 0xD0 :
				_event = new SMFTrackEvent(_deltaTime, _eventType, _channel, bytes.readByte());
				trk = _insertTrackEvent(trk, _event);
				continue;
				
			// System Exclusive ---
				case 0xF0 : 
					_event = new SMFTrackEvent(_deltaTime, 0xF0);
					trk = _insertTrackEvent(trk, _event);
					// Multi Packet Message -->
					if (_multiPacketSysEx)
					{
						// Read SysExData
						_sysExDataSize = _readVarLeng(bytes);
						_event.data = new ByteArray();
						bytes.readBytes(_event.data, 0, _sysExDataSize);
					}
					// Single Packet Message -->
					else
					{
						_skipTime = 0;
						_event.data = new ByteArray();
						_event.data.position = 0;
						//
						while(true)
						{
							_sysExPacketsSize = _readVarLeng(bytes);  // 1-4 byte seaq
							bytes.readBytes(_event.data, _event.data.position, _sysExPacketsSize);
							_event.data.position += _sysExPacketsSize;
							if ((bytes[bytes.position - 1] & 0xff) == 0xF7) break;  // EOX
							_skipTime += _readVarLeng(bytes);
							bytes.position++; // status F7
						}
						_event.data.position = 0;
					}
					continue;
					
				case 0xF7 : 
					if (_multiPacketSysEx)
					{
						_event = new SMFTrackEvent(_deltaTime, 0xF7);
						trk = _insertTrackEvent(trk, _event);
						_sysExDataSize = _readVarLeng(bytes);
						_event.data = new ByteArray();
						bytes.readBytes(_event.data, 0, _sysExDataSize);
					}
					// Non-standard error
					else
					{
						bytes.position = _chunkEnd;
						return false;
					}
					continue;
					
			// Meta Event ---
				case 0xFF : 
					_metaType = bytes.readByte() & 0xff;
					_event = new SMFTrackEvent(_deltaTime, 0xFF00 | _metaType);
					trk = _insertTrackEvent(trk, _event);
					// Read MetaData
					_metaDataSize = _readVarLeng(bytes);
					if (_metaDataSize > 0)
					{
						_event.data = new ByteArray();
						bytes.readBytes(_event.data, 0, _metaDataSize);
					}
					switch (_metaType)
					{
						case 0x2F: bytes.position = _chunkEnd; return true; // End of Track
						case 0x00: _event.value0 = _event.data.readUnsignedShort(); break; // Sequence Number
						case 0x20: _event.value0 = _event.data.readUnsignedByte(); break; // Channel Prefix
						case 0x51: _event.value0 = _event.data.readUnsignedShort() << 8 | _event.data.readUnsignedByte(); break; // Set Tempo
					}
					if (_event.data != null)_event.data.position = 0;
					continue;
			}
		}
		// <-- Finish
		bytes.position = _chunkEnd;
		return true;
	}
	private function _insertTrackEvent(crnt:SMFTrackEvent, ev:SMFTrackEvent):SMFTrackEvent
	{
		if (_format == 0) return crnt.next = ev;
		
		var _next:SMFTrackEvent = crnt.next;
		while (_next != null)
		{
			if (ev.deltaTime >= _next.deltaTime) ev.deltaTime -= _next.deltaTime;
			else
			{
				_next.deltaTime -= ev.deltaTime;
				ev.next = _next;
				return crnt.next = ev;
			}
			crnt = _next;
			_next = _next.next;
		}
		return crnt.next = ev;
	}
	private function _removeUnecEvents(trk:SMFTrackEvent):void
	{
		if (trk == null)return;
		var _eotEv:SMFTrackEvent;
		while (trk.next != null)
		{
			if (trk.next.type == 0xFF2F)
			{
				_eotEv = trk.next;
				if (_eotEv.next == null) return;  // last EOT
				_eotEv.next.deltaTime += _eotEv.deltaTime;
				trk.next = _eotEv.next;
			}
			trk = trk.next;
		}
	}
	private function _readVarLeng(dat:ByteArray):int
	{
		var _rv:int = 0, _v:int = 0;
		do {
			_rv = dat.readByte();
			_v = _v << 7 | _rv & 0x7f;
		} while (_rv & 0x80);
		return _v;
	}
}
}