/*! ********************************************
 * VoiceUnitBase.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth 
{
	
internal class VoiceUnitBase 
{
// Tables
	// Volume, Pan
	protected var _mainVolTable :Vector.<Number> = null;
	protected var _exprVolTable :Vector.<Number> = null;
	protected var _veloVolTable :Vector.<Number> = null;
	protected var _panTable     :Vector.<Number> = null;
// ptr
	internal var process:Function = null;
// Synthesizer Data
	protected var _masterVolume :Number = 0.0;
// VoiceUnit Status
	/*
	- VoiceUnit State -
		         [ FEDC BA98 7654 3210 ] 16Bits Flag.
		value -> [ ---- ---- -KKK dcba ] "-" = Reserved Bit.
		(All Bits "0" = VoiceUnit is DeActive.)
		
		- Operator State -
			         [ 3210 ]
			value -> [ dcba ]
			a = Operator 0
			b = Operator 1
			c = Operator 2
			d = Operator 3
			
		- VoiceUnit Mode - (Exclusive BitFlag)
			         [ FEDC BA98 7654 ]
			value -> [ ---- ---- -KKK ]
			      -> [ ---- ---- -cba ]
			
			a = Key On
			b = Key Off
			c = Sonud Off
	*/
	internal var state      :int = 0;
	internal var partNumber :int = -1;
	internal var noteNumber :int = 0;
	internal var startFrame :int = 0;  // Offset Start Sample Frame
	// Volume
	//protected var _volumeTable	:Vector.<Number> = null;
	protected var _prPan :int = 64;     // preset Pan   (0-127)
	protected var _vPanL :Number = 0.0; // total pan L      (val = _panTable[      _prPan + Part.ccPan])
	protected var _vPanR :Number = 0.0; // total pan R      (val = _panTable[256 - _prPan - Part.ccPan])
	protected var _vPtVol:Number = 0.0; // CC Volume        (val = _mainVolTable[Part.volume])
	protected var _vPtExp:Number = 0.0; // CC Expression    (val = _exprVolTable[Part.expression])
	protected var _vVel  :Number = 0.0; // velocity         (val = _veloVolTable[velocity])
	protected var _vVol  :Number = 0.0; // etc              (val = _vVel * masterVolume)
	protected var _vOut  :Number = 0.0; // output volume    (val = _vPtVol * _vPtExp * _vVol)
	protected var _vOutL :Number = 0.0; // output volume L  (val = _vOut * _vPanL)
	protected var _vOutR :Number = 0.0; // output volume R  (val = _vOut * _vPanR)
	
	internal function killProcess():void
	{
		state      = 0;
		partNumber = -1;
	}
	internal function keyOff():void {}
	internal function soundOff():void {}
	internal function keyOn(ch:int, note:int, velo:int):void {}
	internal function setMasterVolume(val:Number):void
	{
		_masterVolume = val;
		if (state == 0) return;
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	}
	internal function updPartBendSens(val:int):void{}
	internal function updPartPitchBend(val:int):void{}
	internal function updPartModulation(val:int):void{}
	internal function updPartVolume(val:int):void
	{
		_vPtVol = _mainVolTable[val];
		_vOut = _vPtVol * _vPtExp * _vVol;
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	}
	internal function updPartExpression(val:int):void
	{
		_vPtExp = _exprVolTable[val];
		_vOut = _vPtVol * _vPtExp * _vVol;
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	}
	internal function updPartPan(val:int):void
	{
		_vPanL = _panTable[      _prPan + val];
		_vPanR = _panTable[256 - _prPan - val];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	}
// Constructor
	public function VoiceUnitBase():void{}
}
}