/*! ********************************************
 * MidiController.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.util 
{
import flash.utils.ByteArray;
import sora.graybox.GrayBox;
import sora.graybox.IGrayBoxModule;
import sora.graybox.MidiEvent;
import sora.graybox.ProcessData;
public class MidiController implements IGrayBoxModule 
{
// MidiController ---
	public var midiOutPorts :int = 0xFFFF;
	protected var _eventBuffer     :Vector.<MidiEvent> = null;
	protected var _eventBufferSize :int = 64;
	protected var _usedEvents      :int = 0;
	public function clearMIDIEventBuffer():void
	{
		for (var i:int = 0; i < _eventBufferSize; i++)_eventBuffer[i].clear();
		_usedEvents = 0;
	}
	public function getEventBufferSize():int { return _eventBufferSize; }
	public function setEventBufferSize(size:int):int
	{
		if (size == _eventBufferSize) return _eventBufferSize;
		if (size < 16)size = 16;
		_eventBufferSize = size;
		_eventBuffer.fixed = false;
		_eventBuffer.length = size;
		_eventBuffer.fixed = true;
		for (var i:int = 0; i < size; i++)
		{
			if (_eventBuffer[i] == null) _eventBuffer[i] = new MidiEvent();
			_eventBuffer[i].clear();
		}
		_usedEvents = 0;
		return _eventBufferSize;
	}
// MIDI Controlle
	public function noteOff(ch:int, key:int, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0x80;
		_ev.channel = ch;
		_ev.value0 = key;
	}
	public function noteOn(ch:int, key:int, velo:int, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0x90;
		_ev.channel = ch;
		_ev.value0 = key;
		_ev.value1 = velo;
	}
	public function keyAfterTouch(ch:int, key:int, val:int, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xA0;
		_ev.channel = ch;
		_ev.value0 = key;
		_ev.value1 = val;
	}
	public function controlChange(ch:int, ccn:int, val:int, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xB0;
		_ev.channel = ch;
		_ev.value0 = ccn;
		_ev.value1 = val;
	}
	public function programChange(ch:int, num:int, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xC0;
		_ev.channel = ch;
		_ev.value0 = num;
	}
	public function channelAfterTouch(ch:int, val:int, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xD0;
		_ev.channel = ch;
		_ev.value0 = val;
	}
	public function pitchBendChange(ch:int, lsb:int, msb:int, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xE0;
		_ev.channel = ch;
		_ev.value0 = lsb;
		_ev.value1 = msb;
	}
	public function pitchBendChangeW(ch:int, val:int, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xE0;
		_ev.channel = ch;
		_ev.value0 = val & 0x7f;
		_ev.value1 = val >> 7 & 0x7f;
	}
	//** System Exclusive. @param msg ExclusiveMessage @param leng Message Data Length */
	//public function systemExclusive(msg:ByteArray, leng:int, offsetFrames:int = 0):void
	public function systemExclusive(msg:ByteArray, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = 0xF0;
		_ev.channel = 0;
		_ev.data = msg;
	}
	public function midiMessage(type:int, ch:int = 0, val0:int = 0, val1:int = 0, data:ByteArray = null, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = midiOutPorts;
		_ev.offsetFrames = offsetFrames;
		_ev.type = type;
		_ev.channel = ch;
		_ev.value0 = val0;
		_ev.value1 = val1;
		_ev.data = data;
	}
	public function midiMessage_P(portNum:int = 0xffff, type:int = 0x80, ch:int = 0, val0:int = 0, val1:int = 0, data:ByteArray = null, offsetFrames:int = 0):void
	{
		if (_usedEvents >= _eventBufferSize) return;
		var _ev:MidiEvent = _eventBuffer[_usedEvents++];
		_ev.next = null;  // clear
		_ev.midiPorts = portNum & 0xffff;
		_ev.offsetFrames = offsetFrames;
		_ev.type = type;
		_ev.channel = ch;
		_ev.value0 = val0;
		_ev.value1 = val1;
		_ev.data = data;
	}
// Constructor
	public function MidiController(eventBufferSize:int = 64):void
	{
		if (eventBufferSize < 16)eventBufferSize = 16;
		_eventBufferSize = eventBufferSize;
		_eventBuffer = new Vector.<MidiEvent>(_eventBufferSize, true);
		for (var i:int = 0; i < _eventBufferSize; i++) _eventBuffer[i] = new MidiEvent();
		_usedEvents = 0;
	}
// Interface - IfGrayBoxModule -----
	public function updateSampleRate(sampleRate:Number):void{}
	public function updateAudioChannels(ch:int):void{}
	public function updateProcessBlockSize(size:int):void{}
	
	public function resetModule():void{ clearMIDIEventBuffer(); }
	public function initiModule():void{ clearMIDIEventBuffer(); }
	public function connectModule(host:GrayBox):void{ clearMIDIEventBuffer(); }
	public function disconnectModule():void{ clearMIDIEventBuffer(); }
	public function updateBypassStatus(bypass:Boolean):void{ clearMIDIEventBuffer(); }
	
	public function startProcess():void{ clearMIDIEventBuffer(); }
	public function stopProcess():void{ clearMIDIEventBuffer(); }
	public function suspendProcess():void{ clearMIDIEventBuffer(); }
	public function resumeProcess():void{ clearMIDIEventBuffer(); }
	public function process(procData:ProcessData):void
	{
		var i:int = 0, _list:MidiEvent, _ev:MidiEvent;
		for (i = 0; i < _usedEvents; i++)
		{
			_ev = _eventBuffer[i];
			_list = procData.midiEventList;
			while (_list.next != null)
			{
				if (_ev.offsetFrames >= _list.next.offsetFrames) _list = _list.next;
				else break;
			}
			_ev.next = _list.next;
			_list.next = _ev;
		}
		_usedEvents = 0;
	}
}
}